import numbers

from swimlane.exceptions import ValidationError
from .base import Field


class NumberField(Field):

    field_type = (
        'Core.Models.Fields.NumericField, Core',
        'Core.Models.Fields.Numeric.NumericField, Core'
    )

    supported_types = [numbers.Number]

    def __init__(self, *args, **kwargs):
        super(NumberField, self).__init__(*args, **kwargs)

        self.min = self.field_definition.get('min')
        self.max = self.field_definition.get('max')

    def validate_value(self, value):
        super(NumberField, self).validate_value(value)

        if value is not None:
            if self.min is not None and value < self.min:
                raise ValidationError(self.record, "Field '{}' minimum value '{}', received '{}'".format(
                    self.name,
                    self.min,
                    value
                ))

            if self.max is not None and value > self.max:
                raise ValidationError(self.record, "Field '{}' maximum value '{}', received '{}'".format(
                    self.name,
                    self.max,
                    value
                ))



"""Helpers for bulk methods"""


class _BulkModificationOperation(object):
    """Base class for bulk_modify value modification operators

    Acts as container to wrap the modification type with the target value for the bulk operation

    Examples:

        swimlane.records.bulk_modify(
            record,
            values={
                'Field A': 'new value',
                'Field B': Append('new value'),
                'Field C': Clear(),
                ...
            }
        )
    """

    type = None

    def __init__(self, value):
        self.value = value


class Replace(_BulkModificationOperation):
    """Bulk modification 'Replace with'/'Replace all with' operation"""
    type = 'create'


class Clear(_BulkModificationOperation):
    """Bulk modification 'Clear field' operation"""
    type = 'delete'

    def __init__(self):
        super(Clear, self).__init__(None)


class Append(_BulkModificationOperation):
    """Bulk modification 'Add to existing' operation"""
    type = 'append'


class Remove(_BulkModificationOperation):
    """Bulk modification 'Find and remove these' operation"""
    type = 'subtract'



"""Utility functions"""
from __future__ import absolute_import

import importlib
import pkgutil
import random
import string
import functools


def random_string(length, source=string.ascii_letters + string.digits):
    """Return random string of characters from source of specified length

    Args:
        length (int): Length of the returned string
        source (str): String of characters to use as options for randomly selected characters. Defaults to alphanumeric

    Returns:
        str: String of length number of characters composed of source characters
    """
    return ''.join(random.choice(source) for _ in range(length))


def get_recursive_subclasses(cls):
    """Return list of all subclasses for a class, including subclasses of direct subclasses"""
    return cls.__subclasses__() + [g for s in cls.__subclasses__() for g in get_recursive_subclasses(s)]


def import_submodules(package):
    """Return list of imported module instances from beneath root_package"""

    if isinstance(package, str):
        package = importlib.import_module(package)

    results = {}

    for _, full_name, is_pkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        results[full_name] = importlib.import_module(full_name)

        if is_pkg:
            results.update(import_submodules(full_name))

    return results


def one_of_keyword_only(*valid_keywords):
    """Decorator to help make one-and-only-one keyword-only argument functions more reusable

    Notes:
        Decorated function should take 2 arguments, the first for the key, the second the value

    Examples:

        ::

            @one_of_keyword_only('a', 'b', 'c')
            def func(key, value):
                if key == 'a':
                    ...
                elif key == 'b':
                    ...
                else:
                    # key = 'c'
                    ...

            ...

            func(a=1)
            func(b=2)
            func(c=3)

            try:
                func(d=4)
            except TypeError:
                ...

            try:
                func(a=1, b=2)
            except TypeError:
                ...

    Args:
        *valid_keywords (str): All allowed keyword argument names

    Raises:
        TypeError: On decorated call, if 0 or 2+ arguments are provided or kwargs contains a key not in valid_keywords
    """

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            sentinel = object()
            values = {}

            for key in valid_keywords:
                kwarg_value = kwargs.pop(key, sentinel)
                if kwarg_value is not sentinel:
                    values[key] = kwarg_value

            if kwargs:
                raise TypeError('Unexpected arguments: {}'.format(kwargs))

            if not values:
                raise TypeError('Must provide one of {} as keyword argument'.format(', '.join(valid_keywords)))

            if len(values) > 1:
                raise TypeError('Must provide only one of {} as keyword argument. Received {}'.format(
                    ', '.join(valid_keywords),
                    values
                ))

            return func(*(args + values.popitem()))

        return wrapper

    return decorator




from numbers import Number

import six
from shortid import ShortId

from swimlane.core.fields.base import FieldCursor
from swimlane.exceptions import ValidationError
from .base import CursorField

SID = ShortId()


class _ListFieldCursor(FieldCursor):
    """Base class for Text and Numeric FieldCursors emulating a basic list"""

    def _validate_list(self, target):
        """Validate a list against field validation rules"""
        # Check list length restrictions
        min_items = self._field.field_definition.get('minItems')
        max_items = self._field.field_definition.get('maxItems')

        if min_items is not None:
            if len(target) < min_items:
                raise ValidationError(
                    self._record,
                    "Field '{}' must have a minimum of {} item(s)".format(self._field.name, min_items)
                )

        if max_items is not None:
            if len(target) > max_items:
                raise ValidationError(
                    self._record,
                    "Field '{}' can only have a maximum of {} item(s)".format(self._field.name, max_items)
                )

        # Individual item validation
        for item in target:
            self._validate_item(item)

    def _validate_item(self, item):
        """Validate individual item against field rules. Defaults to no-op"""

    def __getattr__(self, item):
        """Fallback to any builtin list methods on self._elements for any undefined methods called on cursor

        List methods are wrapped in a function ensuring the updated value passes field validation before actually
        applying to the internal value

        Wrapper function creates a copy of the current elements in the cursor, calls whatever list method was executed
        on the copy, then validates the updated list against field rules.

        If validation fails, raise ValidationError, otherwise update the cursor elements and field value to the modified
        list copy
        """
        try:
            # Check for method on list class
            func = getattr(list, item)

            # Create a copy of elements if function exists to be the actual target of the method call
            elements_copy = self._elements[:]

            # Wrap in function adding validation after the method is executed
            def wrapper(*args, **kwargs):
                # Execute method against the copied elements
                result = func(elements_copy, *args, **kwargs)

                # Validate elements copy after any potential modification
                self._validate_list(elements_copy)

                # Update internal cursor elements to modified copy and sync with field
                self._elements = elements_copy
                self._sync_field()

                # Return in case of methods retrieving values instead of modifying state
                return result

            return wrapper
        except AttributeError:
            # Raise separate AttributeError with correct class reference instead of list
            raise AttributeError("{} object has no attribute {}".format(self.__class__, item))


class TextListFieldCursor(_ListFieldCursor):
    """Cursor for Text ListField"""

    def _validate_item(self, item):
        """Validate char/word count"""
        if not isinstance(item, six.string_types):
            raise ValidationError(
                self._record,
                "Text list field items must be strings, not '{}'".format(item.__class__)
            )

        words = item.split(' ')

        item_length_type = self._field.field_definition.get('itemLengthType')
        item_max_length = self._field.field_definition.get('itemMaxLength')
        item_min_length = self._field.field_definition.get('itemMinLength')
        if item_length_type is not None:
            # Min/max word count
            if item_length_type == 'words':
                if item_max_length is not None:
                    if len(words) > item_max_length:
                        raise ValidationError(
                            self._record,
                            "Field '{}' items cannot contain more than {} words".format(
                                self._field.name,
                                item_max_length
                            )
                        )
                if item_min_length is not None:
                    if len(words) < item_min_length:
                        raise ValidationError(
                            self._record,
                            "Field '{}' items must contain at least {} words".format(
                                self._field.name,
                                item_min_length
                            )
                        )

            # Min/max char count of full item
            else:
                if item_max_length is not None:
                    if len(item) > item_max_length:
                        raise ValidationError(
                            self._record,
                            "Field '{}' items cannot contain more than {} characters".format(
                                self._field.name,
                                item_max_length
                            )
                        )
                if item_min_length is not None:
                    if len(item) < item_min_length:
                        raise ValidationError(
                            self._record,
                            "Field '{}' items must contain at least {} characters".format(
                                self._field.name,
                                item_min_length
                            )
                        )


class NumericListFieldCursor(_ListFieldCursor):
    """Cursor for Numeric ListField"""

    def _validate_item(self, item):
        if not isinstance(item, Number):
            raise ValidationError(
                self._record,
                "Numeric list field items must be numbers, not '{}'".format(item.__class__)
            )

        # range restrictions
        item_max = self._field.field_definition.get('itemMax')
        item_min = self._field.field_definition.get('itemMin')

        if item_max is not None:
            if item > item_max:
                raise ValidationError(
                    self._record,
                    "Field '{}' items cannot be greater than {}".format(
                        self._field.name,
                        item_max
                    )
                )

        if item_min is not None:
            if item < item_min:
                raise ValidationError(
                    self._record,
                    "Field '{}' items cannot be less than {}".format(
                        self._field.name,
                        item_min
                    )
                )


class ListField(CursorField):
    """Text and Numeric List field"""

    field_type = (
        'Core.Models.Fields.List.ListField, Core',
        'Core.Models.Fields.ListField, Core',
    )

    _type_map = {
        'numeric': {
            'list_item_type': 'Core.Models.Record.ListItem`1[[System.Double, mscorlib]], Core',
            'cursor_class': NumericListFieldCursor
        },
        'text': {
            'list_item_type': 'Core.Models.Record.ListItem`1[[System.String, mscorlib]], Core',
            'cursor_class': TextListFieldCursor
        }
    }

    def __init__(self, *args, **kwargs):
        super(ListField, self).__init__(*args, **kwargs)
        self.cursor_class = self._type_map[self.input_type]['cursor_class']

    def set_swimlane(self, value):
        """Convert from list of dicts with values to list of values"""
        value = [d['value'] for d in value or []]

        return super(ListField, self).set_swimlane(value)

    def set_python(self, value):
        """Validate using cursor for consistency between direct set of values vs modification of cursor values"""
        if not isinstance(value, (list, type(None))):
            raise ValidationError(
                self.record,
                "Field '{}' must be set to a list, not '{}'".format(
                    self.name,
                    value.__class__
                )
            )
        value = value or []
        self.cursor._validate_list(value)
        return super(ListField, self).set_python(value)

    def cast_to_swimlane(self, value):
        value = super(ListField, self).cast_to_swimlane(value)
        return [self._build_list_item(item) for item in value] or None

    def cast_to_bulk_modify(self, value):
        """List fields use raw list values for bulk modify"""
        self.validate_value(value)
        return value

    def _build_list_item(self, item_value):
        """Return a dict with random ID and $type for API representation of value"""
        return {
            '$type': self._type_map[self.input_type]['list_item_type'],
            'id': SID.generate(),
            'value': item_value
        }




import logging

import six
from sortedcontainers import SortedDict

from swimlane.core.fields.base import CursorField, FieldCursor
from swimlane.core.resources.record import Record
from swimlane.exceptions import ValidationError, SwimlaneHTTP400Error

logger = logging.getLogger(__name__)


class ReferenceCursor(FieldCursor):
    """Handles lazy retrieval of target records"""

    def __init__(self, *args, **kwargs):
        super(ReferenceCursor, self).__init__(*args, **kwargs)

        self._elements = self._elements or SortedDict()

    @property
    def target_app(self):
        """Make field's target_app available on cursor"""
        return self._field.target_app

    def _evaluate(self):
        """Scan for orphaned records and retrieve any records that have not already been grabbed"""

        retrieved_records = SortedDict()

        for record_id, record in six.iteritems(self._elements):
            if record is self._field._unset:
                # Record has not yet been retrieved, get it
                try:
                    record = self.target_app.records.get(id=record_id)
                except SwimlaneHTTP400Error:
                    # Record appears to be orphaned, don't include in set of elements
                    logger.debug("Received 400 response retrieving record '{}', ignoring assumed orphaned record")
                    continue

            retrieved_records[record_id] = record

        self._elements = retrieved_records

        return self._elements.values()

    def add(self, record):
        """Add a reference to the provided record"""
        self._field.validate_value(record)
        self._elements[record.id] = record
        self._sync_field()

    def remove(self, record):
        """Remove a reference to the provided record"""
        self._field.validate_value(record)
        del self._elements[record.id]
        self._sync_field()


class ReferenceField(CursorField):

    field_type = 'Core.Models.Fields.Reference.ReferenceField, Core'
    supported_types = (Record,)
    cursor_class = ReferenceCursor

    def __init__(self, *args, **kwargs):
        super(ReferenceField, self).__init__(*args, **kwargs)

        self.__target_app_id = self.field_definition['targetId']
        self.__target_app = None

    @property
    def target_app(self):
        """Defer target app retrieval until requested"""
        if self.__target_app is None:
            self.__target_app = self._swimlane.apps.get(id=self.__target_app_id)

        return self.__target_app

    def validate_value(self, value):
        """Validate provided record is a part of the appropriate target app for the field"""
        if value not in (None, self._unset):

            super(ReferenceField, self).validate_value(value)

            if value.app != self.target_app:
                raise ValidationError(
                    self.record,
                    "Reference field '{}' has target app '{}', cannot reference record '{}' from app '{}'".format(
                        self.name,
                        self.target_app,
                        value,
                        value.app
                    )
                )

    def _set(self, value):
        self._cursor = None
        self._value = value or None

    def set_swimlane(self, value):
        """Store record ids in separate location for later use, but ignore initial value"""

        # Move single record into list to be handled the same by cursor class
        if not self.multiselect:
            if value and not isinstance(value, list):
                value = [value]

        # Values come in as a list of record ids or None
        value = value or []

        records = SortedDict()

        for record_id in value:
            records[record_id] = self._unset

        return super(ReferenceField, self).set_swimlane(records)

    def set_python(self, value):
        """Expect list of record instances, convert to a SortedDict for internal representation"""
        if not self.multiselect:
            if value and not isinstance(value, list):
                value = [value]

        value = value or []

        records = SortedDict()

        for record in value:
            self.validate_value(record)
            records[record.id] = record

        return_value = self._set(records)
        self.record._raw['values'][self.id] = self.get_swimlane()

        return return_value

    def get_swimlane(self):
        """Return list of record ids"""
        value = super(ReferenceField, self).get_swimlane()
        if value:
            ids = list(value.keys())

            if self.multiselect:
                return ids
            else:
                return ids[0]
        else:
            return None

    def get_python(self):
        """Return cursor if multi-select, direct value if single-select"""
        cursor = super(ReferenceField, self).get_python()

        if self.multiselect:
            return cursor

        else:
            try:
                return cursor[0]
            except IndexError:
                return None

    def cast_to_report(self, value):
        return value.id



import weakref

from swimlane.core.cursor import Cursor
from swimlane.core.resolver import SwimlaneResolver
from .field import Field


class FieldCursor(Cursor, SwimlaneResolver):
    """Base class for encapsulating a field instance's complex logic
    
    Useful in abstracting away extra request(s), lazy evaluation, pagination, intensive calculations, etc.
    """

    def __init__(self, field, initial_elements=None):
        SwimlaneResolver.__init__(self, field.record._swimlane)
        Cursor.__init__(self)

        self._elements = initial_elements or self._elements

        self.__field_name = field.name
        self.__record_ref = weakref.ref(field.record)
        self.__field_ref = weakref.ref(field)

    def __repr__(self):
        # pylint: disable=missing-format-attribute
        return '<{self.__class__.__name__}: {self._record!r}["{self._field.name}"] ({length})>'.format(
            self=self,
            length=len(self)
        )

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other._record.id == self._record.id

    def _sync_field(self):
        """Set source field value to current cursor value"""
        self._field.set_python(self._evaluate())

    @property
    def _record(self):
        return self.__record_ref()

    @property
    def _field(self):
        field = self.__field_ref()
        # Occurs when a record is saved and reinitialized, creating new Field instances and losing the existing weakref
        # Update weakref to point to new Field instance of the same name
        if field is None:
            field = self._record.get_field(self.__field_name)
            self.__field_ref = weakref.ref(field)
        return field


class CursorField(Field):
    """Returns a proxy-like FieldCursor instance to support additional functionality"""

    cursor_class = None

    def __init__(self, *args, **kwargs):
        super(CursorField, self).__init__(*args, **kwargs)

        self._cursor = None

    def get_initial_elements(self):
        """Return initial elements to be passed with cursor instantiation"""
        return self._get()

    def _set(self, value):
        self._cursor = None
        super(CursorField, self)._set(value)

    @property
    def cursor(self):
        """Cache and return cursor_class instance"""
        if self._cursor is None:
            # pylint: disable=not-callable
            self._cursor = self.cursor_class(self, self.get_initial_elements())

        return self._cursor

    def get_python(self):
        """Create, cache, and return the appropriate cursor instance"""
        return self.cursor



import copy
from functools import total_ordering

import pendulum
import six

from swimlane.core.resources.base import APIResource
from swimlane.core.resources.usergroup import UserGroup
from swimlane.exceptions import UnknownField, ValidationError


@total_ordering
class Record(APIResource):
    """A single Swimlane Record instance

    Attributes:
        id (str): Full Record ID
        tracking_id (str): Record tracking ID
        created (pendulum.Pendulum): Pendulum datetime for Record created date
        modified (pendulum.Pendulum): Pendulum datetime for Record last modified date
        is_new (bool): True if Record does not yet exist on server. Other values may be temporarily None if True
        app (App): App instance that Record belongs to
    """

    _type = 'Core.Models.Record.Record, Core'

    def __init__(self, app, raw):
        super(Record, self).__init__(app._swimlane, raw)

        self.__app = app

        self.is_new = self._raw.get('isNew', False)

        # Protect against creation from generic raw data not yet containing server-generated values
        if self.is_new:
            self.id = self.tracking_id = self.created = self.modified = None
        else:
            record_app_id = raw['applicationId']
            if record_app_id != app.id:
                raise ValueError('Record applicationId `{}` does not match source app id `{}`'.format(
                    record_app_id,
                    app.id
                ))

            self.id = self._raw['id']

            # Combine app acronym + trackingId instead of using trackingFull raw
            # for guaranteed value (not available through report results)
            self.tracking_id = '-'.join([
                self.app.acronym,
                str(int(self._raw['trackingId']))
            ])

            self.created = pendulum.parse(self._raw['createdDate'])
            self.modified = pendulum.parse(self._raw['modifiedDate'])

        self.__allowed = []

        self._fields = {}
        self.__premap_fields()

    @property
    def app(self):
        return self.__app

    def __str__(self):
        if self.is_new:
            return '{} - New'.format(self.app.acronym)

        return str(self.tracking_id)

    def __setitem__(self, field_name, value):
        self.get_field(field_name).set_python(value)

    def __getitem__(self, field_name):
        return self.get_field(field_name).get_python()

    def __delitem__(self, field_name):
        self[field_name] = None

    def __iter__(self):
        for field_name, field in six.iteritems(self._fields):
            yield field_name, field.get_python()

    def __hash__(self):
        return hash((self.id, self.app))

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("Comparisons not supported between instances of '{}' and '{}'".format(
                other.__class__.__name__,
                self.__class__.__name__
            ))

        tracking_number_self = int(self.tracking_id.split('-')[1])
        tracking_number_other = int(other.tracking_id.split('-')[1])

        return (self.app.name, tracking_number_self) < (other.app.name, tracking_number_other)

    def __premap_fields(self):
        """Build field instances using field definitions in app manifest
        
        Map raw record field data into appropriate field instances with their correct respective types
        """
        # Circular imports
        from swimlane.core.fields import resolve_field_class

        for field_definition in self.app._raw['fields']:
            field_class = resolve_field_class(field_definition)

            field_instance = field_class(field_definition['name'], self)
            value = self._raw['values'].get(field_instance.id)
            field_instance.set_swimlane(value)

            self._fields[field_instance.name] = field_instance

    def get_cache_index_keys(self):
        """Return values available for retrieving records, but only for already existing records"""
        if not (self.id and self.tracking_id):
            raise NotImplementedError

        return {
            'id': self.id,
            'tracking_id': self.tracking_id
        }

    def get_field(self, field_name):
        """Get field instance used to get, set, and serialize internal field value

        Args:
            field_name (str): Field name or key to retrieve

        Returns:
            Field: Requested field instance

        Raises:
             UnknownField: Raised if `field_name` not found in parent App
        """
        try:
            return self._fields[self.app.resolve_field_name(field_name)]
        except KeyError:
            raise UnknownField(self.app, field_name, self._fields.keys())

    def validate(self):
        """Explicitly validate field data

        Notes:
            Called automatically during save call before sending data to server

        Raises:
             ValidationError: If any fields fail validation
        """
        for field in (_field for _field in six.itervalues(self._fields) if _field.required):
            if field.get_swimlane() is None:
                raise ValidationError(self, 'Required field "{}" is not set'.format(field.name))

    def save(self):
        """Persist record changes on Swimlane server
        
        Updates internal raw data with response content from server to guarantee calculated field values match values on
        server
        
        Raises:
            ValidationError: If any fields fail validation
        """

        if self.is_new:
            method = 'post'
        else:
            method = 'put'

        # Pop off fields with None value to allow for saving empty fields
        copy_raw = copy.copy(self._raw)
        values_dict = {}
        for key, value in six.iteritems(copy_raw['values']):
            if value is not None:
                values_dict[key] = value
        copy_raw['values'] = values_dict

        self.validate()

        response = self._swimlane.request(
            method,
            'app/{}/record'.format(self.app.id),
            json=copy_raw
        )

        # Reinitialize record with new raw content returned from server to update any calculated fields
        self.__init__(self.app, response.json())

        # Manually cache self after save to keep cache updated with latest data
        self._swimlane.resources_cache.cache(self)

    def delete(self):
        """Delete record from Swimlane server

        .. versionadded:: 2.16.1

        Resets to new state, but leaves field data as-is. Saving a deleted record will create a new Swimlane record

        Raises
            ValueError: If record.is_new
        """
        if self.is_new:
            raise ValueError('Cannot delete a new Record')

        self._swimlane.request(
            'delete',
            'app/{}/record/{}'.format(self.app.id, self.id)
        )

        del self._swimlane.resources_cache[self]

        # Modify current raw values indicating an unsaved record but persisting field data
        raw = copy.deepcopy(self._raw)
        raw['id'] = None
        raw['isNew'] = True

        self.__init__(self.app, raw)

    @property
    def restrictions(self):
        """Returns cached set of retrieved UserGroups in the record's list of allowed accounts"""
        return [UserGroup(self._swimlane, raw) for raw in self._raw['allowed']]

    def add_restriction(self, *usergroups):
        """Add UserGroup(s) to list of accounts with access to record

        .. versionadded:: 2.16.1

        UserGroups already in the restricted list can be added multiple times and duplicates will be ignored

        Notes:
            Does not take effect until calling `record.save()`

        Args:
            *usergroups (UserGroup): 1 or more Swimlane UserGroup(s) to add to restriction list

        Raises:
            TypeError: If 0 UserGroups provided or provided a non-UserGroup instance
        """
        if not usergroups:
            raise TypeError('Must provide at least one UserGroup for restriction')

        allowed = copy.copy(self._raw.get('allowed', []))

        for usergroup in usergroups:
            if not isinstance(usergroup, UserGroup):
                raise TypeError('Expected UserGroup, received `{}` instead'.format(usergroup))

            selection = usergroup.as_usergroup_selection()
            if selection not in allowed:
                allowed.append(selection)

        self._raw['allowed'] = allowed

    def remove_restriction(self, *usergroups):
        """Remove UserGroup(s) from list of accounts with access to record

        .. versionadded:: 2.16.1

        Notes:
            Does not take effect until calling `record.save()`

        Warnings:
            Providing no UserGroups will clear the restriction list, opening access to ALL accounts

        Args:
            *usergroups (UserGroup): 0 or more Swimlane UserGroup(s) to remove from restriction list

        Raises:
            TypeError: If provided a non-UserGroup instance
            ValueError: If provided UserGroup not in current restriction list
        """
        if usergroups:
            allowed = copy.copy(self._raw.get('allowed', []))

            for usergroup in usergroups:
                if not isinstance(usergroup, UserGroup):
                    raise TypeError('Expected UserGroup, received `{}` instead'.format(usergroup))
                try:
                    allowed.remove(usergroup.as_usergroup_selection())
                except ValueError:
                    raise ValueError('UserGroup `{}` not in record `{}` restriction list'.format(usergroup, self))
        else:
            allowed = []

        self._raw['allowed'] = allowed


def record_factory(app, fields=None):
    """Return a temporary Record instance to be used for field validation and value parsing

    Args:
        app (App): Target App to create a transient Record instance for
        fields (dict): Optional dict of fields and values to set on new Record instance before returning

    Returns:
        Record: Unsaved Record instance to be used for validation, creation, etc.
    """
    # pylint: disable=line-too-long
    record = Record(app, {
        '$type': Record._type,
        'isNew': True,
        'applicationId': app.id,
        'comments': {
            '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.Comments, Core]], mscorlib]], mscorlib'
        },
        'values': {
            '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib'
        }
    })

    fields = fields or {}

    for name, value in six.iteritems(fields):
        record[name] = value

    # Pop off fields with None value to allow for saving empty fields
    copy_raw = copy.copy(record._raw)
    values_dict = {}
    for key, value in six.iteritems(copy_raw['values']):
        if value is not None:
            values_dict[key] = value
    record._raw['values'] = values_dict

    return record



from sortedcontainers import SortedSet

from .cursor import CursorField, FieldCursor


class MultiSelectCursor(FieldCursor):
    """Cursor allowing setting and unsetting values on a MultiSelectField
    
    Respects parent field's validation
    """

    def __init__(self, *args, **kwargs):
        super(MultiSelectCursor, self).__init__(*args, **kwargs)

        self._elements = SortedSet(self._elements)

    def select(self, element):
        """Add an element to the set of selected elements
        
        Proxy to internal set.add and sync field
        """
        self._field.validate_value(element)
        self._elements.add(element)
        self._sync_field()

    def deselect(self, element):
        """Remove an element from the set of selected elements
        
        Proxy to internal set.remove and sync field
        """
        self._elements.remove(element)
        self._sync_field()


class MultiSelectField(CursorField):
    """Base class for fields that can be multi-selection or single-selection field"""

    cursor_class = MultiSelectCursor

    def get_python(self):
        """Only return cursor instance if configured for multiselect"""
        if self.multiselect:
            return super(MultiSelectField, self).get_python()

        return self._get()

    def get_swimlane(self):
        """Handle multi-select and single-select modes"""
        if self.multiselect:
            value = self._get()
            children = []
            if value:
                for child in value:
                    children.append(self.cast_to_swimlane(child))

                return children
            return None
        return super(MultiSelectField, self).get_swimlane()

    def _set(self, value):
        """Override to treat empty lists as None"""
        return super(MultiSelectField, self)._set(value or None)

    def set_python(self, value):
        """Override to remove key from raw data when empty to work with server 2.16+ validation"""
        if self.multiselect:
            value = value or []
            elements = []

            for element in value:
                self.validate_value(element)
                elements.append(element)

            value = elements
        else:
            self.validate_value(value)

        self._set(value)

    def set_swimlane(self, value):
        """Cast all multi-select elements to correct internal type like single-select mode"""
        if self.multiselect:
            value = value or []
            children = []

            for child in value:
                children.append(self.cast_to_python(child))

            return self._set(children)

        return super(MultiSelectField, self).set_swimlane(value)



from swimlane.core.resources.usergroup import UserGroup, User
from swimlane.exceptions import ValidationError
from .base import MultiSelectField


class UserGroupField(MultiSelectField):
    """Manages getting/setting users from record User/Group fields"""

    field_type = (
        'Core.Models.Fields.UserGroupField, Core',
        'Core.Models.Fields.UserGroup.UserGroupField, Core'
    )

    supported_types = [UserGroup]

    def __init__(self, *args, **kwargs):
        super(UserGroupField, self).__init__(*args, **kwargs)

        members = self.field_definition.get('members', [])

        self._allowed_user_ids = set([r['id'] for r in members if r['selectionType'] == 'users'])
        self._allowed_member_ids = set([r['id'] for r in members if r['selectionType'] == 'members'])

        self._allowed_group_ids = set([r['id'] for r in members if r['selectionType'] == 'groups'])
        self._allowed_subgroup_ids = set([r['id'] for r in members if r['selectionType'] == 'subGroups'])

        self._show_all_users = self.field_definition['showAllUsers']
        self._show_all_groups = self.field_definition['showAllGroups']

    def validate_value(self, value):
        """Validate new user/group value against any User/Group restrictions

        Attempts to resolve generic UserGroup instances if necessary to respect special "Everyone" group, and
        "All Users" + "All Groups" options
        """
        super(UserGroupField, self).validate_value(value)

        if value is not None:
            # Ignore validation if all users + groups are allowed
            if self._show_all_groups and self._show_all_users:
                return

            # Try to directly check allowed ids against user/group id first to avoid having to resolve generic
            # UserGroup with an additional request
            if value.id in self._allowed_user_ids | self._allowed_group_ids:
                return

            # Resolve to check Users vs Groups separately
            value = value.resolve()

            if isinstance(value, User):
                self._validate_user(value)
            else:
                self._validate_group(value)

    def _validate_user(self, user):
        """Validate a User instance against allowed user IDs or membership in a group"""
        # All users allowed
        if self._show_all_users:
            return

        # User specifically allowed
        if user.id in self._allowed_user_ids:
            return

        # User allowed by group membership
        user_member_group_ids = set([g['id'] for g in user._raw['groups']])
        if user_member_group_ids & self._allowed_member_ids:
            return

        raise ValidationError(
            self.record,
            'User `{}` is not a valid selection for field `{}`'.format(
                user,
                self.name
            )
        )

    def _validate_group(self, group):
        """Validate a Group instance against allowed group IDs or subgroup of a parent group"""
        # All groups allowed
        if self._show_all_groups:
            return

        # Group specifically allowed
        if group.id in self._allowed_group_ids:
            return

        # Group allowed by subgroup membership
        for parent_group_id in self._allowed_subgroup_ids:
            # Get each group, and check subgroup ids
            parent_group = self._swimlane.groups.get(id=parent_group_id)
            parent_group_child_ids = set([g['id'] for g in parent_group._raw['groups']])
            if group.id in parent_group_child_ids:
                return

        raise ValidationError(
            self.record,
            'Group `{}` is not a valid selection for field `{}`'.format(
                group,
                self.name
            )
        )

    def set_swimlane(self, value):
        """Workaround for reports returning an empty usergroup field as a single element list with no id/name"""
        if value == [{"$type": "Core.Models.Utilities.UserGroupSelection, Core"}]:
            value = []

        return super(UserGroupField, self).set_swimlane(value)

    def cast_to_python(self, value):
        """Convert JSON definition to UserGroup object"""
        # v2.x does not provide a distinction between users and groups at the field selection level, can only return
        # UserGroup instances instead of specific User or Group instances
        if value is not None:
            value = UserGroup(self._swimlane, value)

        return value

    def cast_to_swimlane(self, value):
        """Dump UserGroup back to JSON representation"""
        if value is not None:
            value = value.as_usergroup_selection()

        return value



from swimlane.core.cache import check_cache
from swimlane.core.resolver import SwimlaneResolver
from swimlane.core.resources.app import App
from swimlane.utils import one_of_keyword_only


class AppAdapter(SwimlaneResolver):
    """Handles retrieval of Swimlane App resources"""

    @check_cache(App)
    @one_of_keyword_only('id', 'name')
    def get(self, key, value):
        """Get single app by one of id or name

        Supports resource cache

        Keyword Args:
            id (str): Full app id
            name (str): App name

        Returns:
            App: Corresponding App resource instance

        Raises:
            TypeError: No or multiple keyword arguments provided
            ValueError: No matching app found on server
        """
        if key == 'id':
            # Server returns 204 instead of 404 for a non-existent app id
            response = self._swimlane.request('get', 'app/{}'.format(value))
            if response.status_code == 204:
                raise ValueError('No app with id "{}"'.format(value))

            return App(
                self._swimlane,
                response.json()
            )
        else:
            # Workaround for lack of support for get by name
            # Holdover from previous driver support, to be fixed as part of 3.x
            for app in self.list():
                if value and value == app.name:
                    return app

            # No matching app found
            raise ValueError('No app with name "{}"'.format(value))

    def list(self):
        """Retrieve list of all apps

        Returns:
            :class:`list` of :class:`~swimlane.core.resources.app.App`: List of all retrieved apps
        """
        response = self._swimlane.request('get', 'app')
        return [App(self._swimlane, item) for item in response.json()]



import six

from .base import Field


class TextField(Field):

    field_type = (
        'Core.Models.Fields.TextField, Core',
        'Core.Models.Fields.Text.TextField, Core'
    )

    supported_types = six.string_types

    def set_python(self, value):
        """Set field internal value from the python representation of field value"""
        
        # hook exists to stringify before validation

        # set to string if not string or unicode
        if value is not None and not isinstance(value, self.supported_types) or isinstance(value, int):
            value = str(value)
        return super(TextField, self).set_python(value)

